//: /be/optiloading/Chromosome.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading;

import java.io.*;
import java.util.ArrayList;
import java.util.Collections;

/** 
 * Chromosome class which stores properties of each chromosome and handles methods as creation, mutation and crossover.
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank" href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class Chromosome implements Comparable<Chromosome>, Cloneable, Serializable{
	private float fitness;
	private ArrayList<Float> solution;
	private ArrayList<Float> volumes;
	
	/**
	 * Constructor which creates a new Chromosome using the full tank method.
	 * The full tank method fills tanks completely until one tank gets what left over.
	 */
	public Chromosome(){
		Ship usedship = Ship.getInstance(); // Get a Ship Object
		float toload = usedship.toLoad.getVolume(); // Get the to be loaded volume
		volumes = new ArrayList<Float>(); // Create a new ArrayList to stow the volumes
		ArrayList<Integer> randtank = new ArrayList<Integer>(); // Create an ArrayList to stow a list of tank numbers (to randomly select a tank and then delete the member; every tank will only be selected once)
		
		// Fill the list of tanknumbers and set the default volumes to 0
		for(int i=0;i<usedship.tanks.size();i++){
			randtank.add(i);
			volumes.add(0f);
		}
		
		// Itterate for the number of tanks
		for(int i=0;i<usedship.tanks.size();i++){
			int random = (int) Math.round((Math.random()*(randtank.size()-1))); //Get a random index value between 0 and the number of tanks in the tanklist -1
			int rand = randtank.get(random); //Get the number of the tank from the list
			Tank presenttank = usedship.tanks.get(rand); // Set the present tank to the tank with that index value
			float volume = (float) presenttank.getMaxvolume(); //Set the volume to the maximum volume of that tank
			
			// Check if the to be loaded value is not exceeded else set the volume to the to be loaded value and break
			if (toload < volume){
				volume = toload;
				volumes.set(rand,volume);
				toload -= volume;
				break;
			}
			volumes.set(rand,volume); // Set the volume value in the list to the correct value
			toload -= volume; // Subtract the volume from the to be loaded value
			randtank.remove(random); // Remove the tank number from the list with tank numbers
		}
		usedship = null;
		evaluateFitness(); // Evaluate the fitness
	}
	
	/**
	 * Constructor which creates a new Chromosome using the random tank method.
	 * The random tank method fills tanks totally random.
	 * @param chose Boolean value indicating that u want to use this method to create new chromosomes.
	 */
	public Chromosome(boolean chose){
		Ship usedship = Ship.getInstance(); // Get a Ship object
		float toload = usedship.toLoad.getVolume(); // Get the to be loaded volume
		volumes = new ArrayList<Float>(); // Initialize new ArrayList to stow the volume values
		
		// Set the standard volume values to zero
		for(int i=0;i<usedship.tanks.size();i++){
			volumes.add(0f);
		}
		
		// Distibute volumes until there is nothing left to distribute
		while(toload != 0){
			int rand = (int) Math.round((Math.random()*(usedship.tanks.size()-1))); // Get a index value from 0 to to the number of tanks - 1
			Tank presenttank = usedship.tanks.get(rand); // Assign the tank at the random index value to the presenttank object
			float volume = (float) Math.random()*(presenttank.getMaxvolume()-volumes.get(rand)); // Get a random value between 0 and the space left in the tank
			
			// Check if the to be loaded value is not exceeded else set the volume to the to be loaded value and break
			if (toload < volume){
				volume = volumes.get(rand) + toload;
				volumes.set(rand,volume);
				toload -= volume;
				break;
			}
			volumes.set(rand,(volumes.get(rand)+volume)); //Add the volume to the current volume value
			toload -= volume; // Subtract the volume from the to be loaded volume
		}
		usedship = null;
		evaluateFitness(); //Evaluate fitness
	}
	
	/**
	 * compareTo method is the overwritten method of the Object object which tells how two Chromosomes should be compared.
	 * This method is used to sort the Chromosome ArrayList in the population to get the best Chromosome and to use rank selection.
	 * @param c Chromosome object which is to be compared to this Chromosome object.
	 */
	public int compareTo(Chromosome c){
		return Float.valueOf(fitness).compareTo(c.getFitness()); //Chromosomes are compared by their fitness value
	}

	/**
	 * eveluateFitness is the method used for setting the fitness value of the current Chromosome.
	 */
	private void evaluateFitness(){
		float fsfactor = 0;
		float trimfactor = 0;
		float strengthfactor = 0;
		
		//Calculate the solutions
		Ship usedShip = Ship.getInstance();
		solution = new ArrayList<Float>(usedShip.getSolution(volumes));
		float cos = (float) Math.cos(Math.PI/6);
		float dynstab = (float) (-getGM()*cos + 0.5*getBM()*(cos + (1/cos)) + getGM() -getBM());
		solution.add(dynstab);
		
		
		//Free Surface Effect (20-80) rule
		float maxvalue = 1f/(2*volumes.size()); // Max factor value of each tank
		
		// Iterate for every tank
		for(int i = 0; i < volumes.size(); i++){
			float percentage = volumes.get(i)/usedShip.tanks.get(i).getMaxvolume(); // Get the percentage of how full the tank is
			
			// Below 20 %
			if((percentage<=0.2f) && (percentage != 0)){
				fsfactor += (maxvalue * ( Math.pow(((0.2f-percentage)/0.2f),2) )); // Increment the factor value by the max factor value of each tank multiplied by a factor which increases when the tank is less full 
			}
			
			// Above 80 %
			if((percentage>=0.8f) && (percentage != 1)){
				fsfactor += (maxvalue * ( Math.pow(((percentage-0.8f)/0.2f), 2) )); // Increment the factor value by the max factor value of each tank multiplied by a factor which increases when the tank is more full 
			}
			
			// 100 % or 0 %
			if((percentage == 1f) || (percentage == 0f)){
				fsfactor += (3*maxvalue);
			}
		}
		
		// Trim
		float trimround1 = solution.get(0)*1000;
		float trimround2 = Math.round(trimround1);
		float trimrounded = (float) (trimround2/(1000));

		if ((Genetics.TARGET) != trimrounded){ // Division by zero would lead to errors though is not probable
			trimfactor = (float) (2*(1f /Math.sqrt((Math.abs(Genetics.TARGET - trimrounded))))); // Fitness value defined by the absolute value of 1/(Target-Current)
		} else {trimfactor=100;}
		
		
		// Strength
		ArrayList<Intensity> newintensitylist = new ArrayList<Intensity>();
		
		//Get a clone of the ship's intensity list
		for(Intensity x: usedShip.intensitylist){
			newintensitylist.add((Intensity) x.clone());
		}
		float specgrav = usedShip.toLoad.getSpecificgravity();

		//Add each tanks intensity to the list
		int j=0;
		for(Tank x: usedShip.tanks){
			float p = (float) ((volumes.get(j)*specgrav)/(-x.getLength()));
			for(int i = (x.getBeginframe()+1); i < (x.getEndframe()+1); i++){
				newintensitylist.get(i).addIntensity(p);
			}
			j++;
		}
		
		//Calculate the mean value of the absolute values of the intensity
		//The lower this value the less shear forces and bending moments
		float totalintensity = 0;
		for(Intensity x: newintensitylist){
			totalintensity += Math.abs(x.getIntensity());
		}
		float mean = totalintensity/247;
		
		strengthfactor = 100000/mean;
		
		//Check if the strength criteria is checked in the settings panel
		if(Genetics.STRENGTH == false){
			strengthfactor = 1;
		}
		
		//Check if the trim criteria is checked in the settings panel
		if(Genetics.TRIM == false){
			trimfactor = 1;
		}
		
		//Check if the stability criteria is checked in the settings panel
		if(Genetics.STABILITY == false){
			fsfactor = 1;
		}
		
		fitness = strengthfactor * trimfactor * fsfactor; //Calculate the fitness value
		
		usedShip = null;
	}
	
	/**
	 * getFitness method is used to get the fitness value of the current Chromosome.
	 * @return Floating point value indicating the fitness of the Chromosome.
	 */
	public float getFitness(){
		return fitness;
	}
	
	/**
	 * getVolumes method is used to get the ArrayList of the volume values of the current Chromosome.
	 * @return ArrayList which contains floating point values that indicate the volume of each tank.
	 */
	public ArrayList<Float> getVolumes(){
		return volumes;
	}
	
	/**
	 * getTrim method is used to get the trim value of the current Chromosome.
	 * @return Floating point value indicating the trim of the Chromosome.
	 */
	public float getTrim(){
		return solution.get(0);
	}
	
	/**
	 * getDfwd method is used to get the draught fwd of the current Chromosome.
	 * @return Floating point value indicating the draught fwd of the Chromosome.
	 */
	public float getDfwd(){
		return solution.get(1);
	}
	
	/**
	 * getDaft method is used to get the draught aft of the current Chromosome.
	 * @return Floating point value indicating the draught aft of the Chromosome.
	 */
	public float getDaft(){
		return solution.get(2);
	}
	
	/**
	 * getGM method is used to get the GM value of the current Chromosome.
	 * @return Floating point value indicating the GM of the Chromosome.
	 */
	public float getGM(){
		return solution.get(3);
	}
	
	/**
	 * getFSC method is used to get the free surface correction of the current Chromosome.
	 * @return Floating point value indicating the free surface correction of the Chromosome.
	 */
	public float getFSC(){
		return solution.get(4);
	}
	
	/**
	 * getBM method is used to get the BM of the current Chromosome.
	 * @return Floating point value indicating the BM of the Chromosome.
	 */
	public float getBM(){
		return solution.get(5);
	}
	
	/**
	 * getDynstab method is used to get the area under 0-30 of the current Chromosome.
	 * @return Floating point value indicating the area under 0-30 of the Chromosome.
	 */
	public float getDynstab(){
		return solution.get(6);
	}
	
	/**
	 * crossOverMean method is used to perform a crossover where the the mean value of each tank volume is taken.
	 * @param other Chromosome object to crossover with this Chromosome object.
	 */
	public void crossOverMean(Chromosome other){
		if(Math.random()>=Genetics.CROSSRATE){return;} //Should there be crossover depending on the crossover rate (how lower this rate, how higher the chance that the random value is higher)
		
		// Iterate trough all the tanks
		for(int i = 0; i < volumes.size(); i++){
			float mean = (volumes.get(i)+other.getVolumes().get(i))/2; // Get the mean value
			volumes.set(i,mean); // Set the mean value
		}
		evaluateFitness(); // Evaluate the fitness
	}
	
	/**
	 * crossOverOnePoint method is used to perform a crossover where until a certain point we get volumes from this chromosome and from that point on combine them with volumes from the other chromosome.
	 * @param other Chromosome object to crossover with this Chromosome object.
	 */
	public void crossOverOnePoint(Chromosome other){
		if(Math.random()>=Genetics.CROSSRATE){return;} //Should there be crossover depending on the crossover rate (how lower this rate, how higher the chance that the random value is higher)
		
		Ship usedShip = Ship.getInstance(); // Get a Ship object
		float toload = usedShip.toLoad.getVolume(); // Get the to be loaded volume
		 
		ArrayList<Float> original = new ArrayList<Float>(volumes); // Create an ArrayList for the original volumes (the volumes are changed and might not be valid)
		
		// Do while the solution is not valid
		do{
			
			float rand = (float) Math.random()*toload; // Get a random value between 0 and the to be loaded volume (crossover point)
			float rand2 = toload - rand; //Get the complement of rand: the volume that has to be taken from the other Chromosome
			volumes = original; // Set the volumes back to the original when the new volumes weren't valid
			
			// Get the volumes from this Chromosome until the crossover point starting at the first tank
			for(int i = 0; i < volumes.size(); i++){
				float vol = volumes.get(i);
				if(vol>rand){
					vol = rand;
				}
				rand -= vol; 
				volumes.set(i,vol);
			}
			
			// Get the volumes from the other Chromosome starting from the last tank and add them to the volumes of this Chromosome
			for(int i = (volumes.size()-1); i > -1; i--){
				float vol = other.getVolumes().get(i);
				if(vol>rand2){
					vol = rand2;
				}
				rand2 -= vol;
				vol += volumes.get(i);
				volumes.set(i,vol);
			}
		} while (checkChrom() == false); // Check if the new volumes are valid
		original = null;
		usedShip = null;
		evaluateFitness(); // Evaluate fitness
	}
	
	/**
	 * crossOverTwoPoint method is used to perform a crossover where we take volumes from the other chromosome from one point to another and put those volumes in between the values of the original chromosome.
	 * @param other Chromosome object to crossover with this Chromosome object.
	 */
	public void crossOverTwoPoint(Chromosome other){
		if(Math.random()>=Genetics.CROSSRATE){return;}
		
		Ship usedShip = Ship.getInstance();
		float toload = usedShip.toLoad.getVolume();
		
		ArrayList<Float> original = new ArrayList<Float>(volumes); // Create an ArrayList for the original volumes (the volumes are changed and might not be valid)
		ArrayList<Float> copy = new ArrayList<Float>(volumes); // Create a copy of the volumes Arraylist
		
		do{
			
			float rand1 = (float) Math.random()*toload; //Get the random value to be taken from this Chromosome
			float rand2 = (float) Math.random()*(toload-rand1); // Get the randomv alue to be taken from the other Chromosome
			float rand3 = toload-rand1-rand2; // Get the leftover value which is to be taken from this Chromosome
			float copyrand1 = rand1;
			
			// Set the values back to their original values
			volumes = original;
			ArrayList<Float> copyother = new ArrayList<Float>(other.getVolumes());
			
			// Take the rand1 from this Chromosome starting at the first tank
			for(int i = 0; i < copy.size(); i++){
				float vol = copy.get(i);
				if(vol>rand1){
					vol = rand1;
				}
				rand1 -= vol; // If the previous if-statement is true this value becomes zero and all following volumes become zero seeing vol>rand1 will always be true
				volumes.set(i,vol);
			}
			
			// Take the rand3 from this Chromosome starting at the last tank
			for(int i = (volumes.size()-1); i > -1; i--){
				float vol = copy.get(i);
				if(vol>rand3){
					vol = rand3;
				}
				rand3 -= vol;
				vol += volumes.get(i);
				volumes.set(i,vol);
			}
			
			// Subtract rand1 from the other Chromosome starting at the first tank
			for(int i = 0; i < copyother.size(); i++){
				float vol = copyother.get(i);
				if(vol>copyrand1){
					vol -= copyrand1;
					copyrand1 = 0;
				} else {
					copyrand1 -= vol;
					vol =0;
				}
				copyother.set(i,vol);
			}
			
			// Take rand2 from the other Chromosome starting at the first tank and add it to the volumes of this array
			for(int i = 0; i < copyother.size(); i++){
				float vol = copyother.get(i);
				if(vol>rand2){
					vol = rand2;
				}
				rand2 -= vol;
				vol += volumes.get(i);
				volumes.set(i,vol);
			}
			
		} while (checkChrom() == false); // Check if the new volumes are valid
		original = null;
		usedShip = null;
		evaluateFitness(); // Evaluate the fitness
	}
	
	/**
	 * crossOverUniform method is used to perform a crossover where the volume of a tank is randomly taken from this or the other chromosome, after which a correction is applied.
	 * @param other Chromosome object to crossover with this Chromosome object.
	 */
	public void crossOverUniform(Chromosome other){
		if(Math.random()>=Genetics.CROSSRATE){return;} //Should there be crossover depending on the crossover rate (how lower this rate, how higher the chance that the random value is higher)
		
		do {
			float sum = 0; // Initializes the value that is used to make the sum of all the selected volumes (to later apply the correction)
			ArrayList<Float> original = new ArrayList<Float>(volumes); // Create an ArrayList for the original volumes (the volumes are changed and might not be valid)
			while(sum == 0){ // To prevent a case where every volume is equal to zero (the correction would not work in that case)
				volumes = original; // Set the volumes back to the orignal values if the new solution isn't valid
				
				// Iterate trough the volumes
				for(int i = 0; i < volumes.size();i++){
					if(Math.round(Math.random()) == 1){ // There's a 50-50 chance that the volume of the other Chromosome will be used.
						float othervol = other.getVolumes().get(i);
						volumes.set(i,othervol); // Set the volume to the value of the other tank
						sum += othervol; //Add to the sum of all volumes
					} else {sum += volumes.get(i);}
				}
			}
			
			Ship usedShip = Ship.getInstance();
			float correction = usedShip.toLoad.getVolume() - sum; // The correction is the difference between the to be loaded value and the sum of the already loaded volumes (can either be positive or negative)
			
			// Itereate trough the tanks applying the correction while preserving the ratios
			for(int i = 0; i < volumes.size();i++){
				float newvol = ((volumes.get(i)/sum)*correction)+volumes.get(i);
				volumes.set(i,newvol);
			}
		} while (checkChrom() == false); // Check if the new volumes are valid
		
		evaluateFitness(); // Evaluate the fitness
	}
	
	/**
	 * mutateSwitch method randomly switches tank volumes and checks if the new solution is valid.
	 */
	public void mutateSwitch(){
		if(Math.random()>=Genetics.MUTRATE){return;} //Should there be mutation depending on the mutation rate (how lower this rate, how higher the chance that the random value is higher)
		
		do{
			Collections.shuffle(volumes); // Shuffle volumes
		} while (checkChrom() == false); // Check if new volumes are valid
		
		evaluateFitness();
	}
	
	/**
	 * mutateRandom method subtracts a random value from one tank and add this value to another tank.
	 */
	public void mutateRandom(){
		if(Math.random()>=Genetics.MUTRATE){return;} //Should there be mutation depending on the mutation rate (how lower this rate, how higher the chance that the random value is higher)
		
		// Get ship object and initialize primitives
		Ship usedShip = Ship.getInstance();
		int rand1 = 0;
		int rand2 = 0;
		float randomvol = 0;
		ArrayList<Float> original = new ArrayList<Float>(volumes); // Create an ArrayList for the original volumes (the volumes are changed and might not be valid)
		
		do{
			volumes = original; // Set the volumes back to their original values
			
			// Get a random tank that is not empty
			do{
				rand1 = (int) Math.round((Math.random()*(volumes.size()-1)));
			} while(volumes.get(rand1) == 0f);

			// Get a random volume from that tank and subtract that value
			randomvol = (float) Math.random()*volumes.get(rand1);
			float newvol1 = volumes.get(rand1)-randomvol;
			volumes.set(rand1, newvol1);
			
			// Get a second random tank that is not the same tank as the first
			do{
				rand2 = (int) Math.round((Math.random()*(volumes.size()-1)));
			} while(rand2 == rand1);
			
			// Add the previously subtracted value to this new tank
			float newvol2 = volumes.get(rand2)+randomvol;
			volumes.set(rand2, newvol2);
			
		} while (checkChrom()==false); // Check if the new volumes are valid.

		evaluateFitness(); // Evaluate fitness
	}
	
	// Check if the Chromosome is valid (not exceeding max tank volumes)
	private boolean checkChrom(){
		boolean returnvalue = true; // Initialize return value
		Ship usedShip = Ship.getInstance(); // Get Ship object
		
		// Iterate trough all volumes and check if they are not exceeding the max tank volume and are not negatieve (this could happen when using the crossOverUniform method)
		for(int i = 0; i < volumes.size(); i++){
			if((volumes.get(i)> usedShip.tanks.get(i).getMaxvolume()) ||
			(volumes.get(i)<0f)){
				returnvalue = false;
				break;
			}
		}
		usedShip = null;
		return returnvalue;
	}
	
	/**
	 * clone method is the overwritten method of the Object object which also copies the ArrayList of volumes.
	 * @return Object object which is a clone of the current Chromosome object and should be cast as a Chromosome.
	 */
	public Object clone(){
		try {
			Object clone = super.clone(); // Make clone
			((Chromosome) clone).volumes = new ArrayList<Float>(volumes); // Copy volumes (the reason this method is overwritten)
			return clone;
		} catch (CloneNotSupportedException e){
			throw new InternalError();
		}
	}
}