//: /be/optiloading/Population.java
/*
 * This work is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This work is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * Copyright (c) 2007 Tim Pintens. All rights reserved.
 */

package be.optiloading;

import java.util.ArrayList;
import java.util.Collections;

/** 
 * Population class stores Chromosomes and selects them.
 * @author Tim Pintens, info@optiloading.be
 * @author <a target="_blank" href="http://www.optiloading.be">www.optiloading.be</a>
 */

public class Population {
	private ArrayList<Chromosome> pop; //ArrayList of Chromosomes
	private float totalfitness;
	
	/**
	 * Construtor initializes the Arraylist
	 */
	public Population(){
		pop = new ArrayList<Chromosome>();
	}
	
	/**
	 * Creates a new population; keeping into account the elitism rate.
	 * @param rate Integer value indicating how many chromosomes to copy from original population.
	 */
	public Population elitism(int rate){
		Population returnvalue = new Population();
		for(int i = 0; i < rate; i++ ){
			int j = Genetics.POPSIZE - 1;
			Chromosome newchrom = (Chromosome) pop.get(j).clone();
			returnvalue.addChromosome(newchrom);
			j--;
		}
		return returnvalue;
	}
	
	/**
	 * creatInitialPopulation method creates the initial population
	 */
	public void createInitialPopulation(){
		Chromosome newchrom = null;
		
		// Itteration as many times as the population size
		for(int i = 0; i < Genetics.POPSIZE;i++){
			
			switch(Genetics.CREATION){
				case 0: newchrom = new Chromosome(true); break; //F ill tanks completely
				case 1: newchrom = new Chromosome(); break;     // Randomly fill tanks
			}
			totalfitness += newchrom.getFitness(); //Incriment the total fitness
			pop.add(newchrom); // Add the Chromosome to the population
		}
	}
	
	/**
	 * addChromosome method adds a new Chromosome to the population.
	 * @param chrom Chromosome object that is to be added.
	 */
	public void addChromosome(Chromosome chrom){
		pop.add(chrom);
		totalfitness += chrom.getFitness();
		Collections.sort(pop); // Sort the population from Chromosome with small fitness values to Chromosomes with large fitness values
	}
	
	/**
	 * getTotalfitness method for getting the total fitness.
	 * @return Floating point value containing the total fitness.
	 */
	public float getTotalfitness(){
		return totalfitness;
	}
	
	/**
	 * getBestchromosome method for getting the best Chromosome.
	 * @return Chromosome object which is the best of its population
	 */
	public Chromosome getBestchromosome(){
		return pop.get(Genetics.POPSIZE-1);
	}
	
	/**
	 * getChromList method for getting the list of Chromosome.
	 * @return ArrayList containing the Chromosome objects.
	 */
	public ArrayList<Chromosome> getChromList(){
		return pop;
	}
	
	/**
	 * selectRouletteChromosome method uses the roulette selection for selecting a Chromosome.
	 * @return Chromosome object that is chosen.
	 */
	public Chromosome selectRouletteChromosome(){
		
		// Get a random value between 0 and the total fitness
		float random = (float) Math.random()*totalfitness;
		
		// Initialize the floating point value that will hold the sum of the different fitness value until this sum is larger then the total value
		float sum = 0;
		
		// Initialize return value
		Chromosome returnvalue = null;
		
		// Iterate trough the population until the sum is larger then the random value
		for(Chromosome chrom: pop){
			sum += chrom.getFitness();
			if(sum >= random){
				returnvalue = (Chromosome) chrom.clone(); // Return a clone so as not to change the original population (total fitness would change)
				break;
			}
		}
		return returnvalue;
	}
	
	/**
	 * selectRankChromosome method uses the rank selection for selecting a Chromosome.
	 * Chromosomes are order according to their fitness value from small to large.
	 * The index number of each Chromosome is used to select the Chromosome. 
	 * @return Chromosome object that is chosen.
	 */
	public Chromosome selectRankChromosome(){
		int sum = ((0+pop.size())*(pop.size()-1))/2; // Get the sum of the index numbers (depending on the size of the population)
		int random = (int) Math.round(Math.random()*sum); // Get a random integer value between 0 and the sum of the index numbers
		sum = 0; // Set the sum back to 0 to begin the iteration
		Chromosome returnvalue = null; // Initialize the return value
		
		// Iterate trough the population until the sum of the index numbers is larger then the random value
		for(int i = 0; i < pop.size(); i++){
			sum += i;
			if(sum >= random){
				returnvalue = (Chromosome) pop.get(i).clone(); // Return a clone so as not to change the original population (total fitness would change)
				break;
			}
		}
		return returnvalue;
	}
	
}